unit Q3BSP;

//------------------------------------------------------------------------
//
// Author      : Jan Horn
// Email       : jhorn@global.co.za
// Website     : http://www.sulaco.co.za
//               http://home.global.co.za/~jhorn
// Date        : 2 June 2002
// Version     : 1.0
// Description : Loads faces and textures from a Quake3 BSP file
//               Loads and renders lightmaps for the level
//
//------------------------------------------------------------------------
// Modifyed By Stucuk To Work With Sascha Willems's Files
//


interface

Uses dglOpenGL, Windows, GLMath;

  const FACE_POLYGON = 1;
  const MAX_TEXTURES = 1000;
  var Gamma : glFloat;

  type TBSPHeader = Record
    strID : Array[0..3] of Char;           // This should always be 'IBSP'
    Version : Integer;                     // This should be 0x2e for Quake 3 files
  end;

  type TBSPLump = Record
    Offset : Integer;                      // The offset into the file for the start of this lump
    Length : Integer;                      // The length in bytes for this lump
  end;

  type TBSPVertex = Record
    Position      : TVector3f;             // (x, y, z) position.
    TextureCoord  : TVector2f;             // (u, v) texture coordinate
    LightmapCoord : TVector2f;             // (u, v) lightmap coordinate
    Normal        : TVector3f;             // (x, y, z) normal vector
    Color         : Array[0..3] of Byte    // RGBA color for the vertex
  end;

  type TBSPFace = Record
    textureID : Integer;                   // The index into the texture array
    effect    : Integer;                   // The index for the effects (or -1 = n/a)
    FaceType  : Integer;                   // 1=polygon, 2=patch, 3=mesh, 4=billboard
    startVertIndex : Integer;              // The starting index into this face's first vertex
    numOfVerts     : Integer;              // The number of vertices for this face
    meshVertIndex  : Integer;              // The index into the first meshvertex
    numMeshVerts   : Integer;              // The number of mesh vertices
    lightmapID     : Integer;              // The texture index for the lightmap
    lMapCorner : Array[0..1] of Integer;   // The face's lightmap corner in the image
    lMapSize   : Array[0..1] of Integer;   // The size of the lightmap section
    lMapPos  : TVector3f;                  // The 3D origin of lightmap.
    lMapVecs : Array[0..1] of TVector3f;   // The 3D space for s and t unit vectors.
    vNormal  : TVector3f;                  // The face normal.
    Size : Array[0..1] of Integer;         // The bezier patch dimensions. 
  end;

  type TBSPTexture = Record
    TextureName : Array[0..63] of Char;    // The name of the texture w/o the extension
    flags    : Integer;                    // The surface flags (unknown)
    contents : Integer;                    // The content flags (unknown)
  end;

  type TBSPLightmap = Record
    imageBits : Array[0..127, 0..127, 0..2] of Byte;   // The RGB data in a 128x128 image
  end;

  type TQuake3BSP = Object
    private

    public
      numOfVerts     : Integer;             // The number of verts in the model
      numOfFaces     : Integer;             // The number of faces in the model
      numOfTextures  : Integer;             // The number of texture maps
      numOfLightmaps : Integer;             // The number of light maps
      Vertices       : Array of TBSPVertex; // The object's vertices
      Faces          : Array of TBSPFace;   // The faces information of the object

      // The texture and lightmap array for the level
      Textures  : Array of glUint;
      procedure RenderFace(faceIndex : Integer);
      function  LoadBSP(const FileName : String) : Boolean;
      procedure RenderLevel(const Pos : TVector3f);
  end;

  // This is our lumps enumeration
  const
    kEntities    = 0;            // Stores player/object positions, etc...
    kTextures    = 1;            // Stores texture information
    kPlanes      = 2;            // Stores the splitting planes
    kNodes       = 3;            // Stores the BSP nodes
    kLeafs       = 4;            // Stores the leafs of the nodes
    kLeafFaces   = 5;            // Stores the leaf's indices into the faces
    kLeafBrushes = 6;            // Stores the leaf's indices into the brushes
    kModels      = 7;            // Stores the info of world models
    kBrushes     = 8;            // Stores the brushes info (for collision)
    kBrushSides  = 9;            // Stores the brush surfaces info
    kVertices    = 10;           // Stores the level vertices
    kMeshVerts   = 11;           // Stores the model vertices offsets
    kShaders     = 12;           // Stores the shader files (blending, anims..)
    kFaces       = 13;           // Stores the faces for the level
    kLightmaps   = 14;           // Stores the lightmaps for the level
    kLightVolumes= 15;           // Stores extra world lighting information
    kVisData     = 16;           // Stores PVS and cluster info (visibility)
    kMaxLumps    = 17;           // A constant to store the number of lumps


implementation

Uses SysUtils, Textures;

{ TQuake3BSP }

{-----------------------------------------------------------}
{---  This loads in all of the .bsp data for the level   ---}
{-----------------------------------------------------------}
function TQuake3BSP.LoadBSP(const FileName: String): Boolean;
var F : File;
    I : Integer;
    Temp : glFloat;
    Header : TBSPHeader;
    Lumps  : Array of TBSPLump;
    BSPTextures  : Array of TBSPTexture;
begin
  result :=FALSE;

  // Check if the .bsp file can be opened
  AssignFile(F, filename);
{$I-}
  Reset(F,1);
{$I+}
  if IOResult <> 0 then
  begin
    MessageBox(0, 'Could not find the BSP file!', 'Error', MB_OK);
    exit;
  end;

  SetLength(Lumps, kMaxLumps);

  // Read in the header and lump data
  BlockRead(F, Header, Sizeof(Header));
  BlockRead(F, Lumps[0], kMaxLumps*sizeof(TBSPLump));

  // Allocate the vertex memory
  numOfVerts := Round(lumps[kVertices].length / sizeof(TBSPVertex));
  SetLength(Vertices, numOfVerts);

  // Allocate the face memory
  numOfFaces := Round(lumps[kFaces].length / sizeof(TBSPFace));
  SetLength(Faces, numOfFaces);

  // Allocate memory to read in the texture information.
  numOfTextures := Round(lumps[kTextures].length / sizeof(TBSPTexture));
  SetLength(BSPTextures, numOfTextures);
  SetLength(Textures, numOfTextures);

  // Seek to the position in the file that stores the vertex information
  Seek(F, lumps[kVertices].offset);

  // Go through all of the vertices that need to be read and swap axises
  for I :=0 to numOfVerts-1 do
  begin
    // Read in the current vertex
    BlockRead(F, Vertices[i], sizeOf(TBSPVertex));

    // Swap the y and z values, and negate the new z so Y is up.
    Temp := Vertices[i].Position.Y;
    Vertices[i].Position.Y := Vertices[i].Position.Z;
    Vertices[i].Position.Z := -temp;

    // Negate the V texture coordinate because it is upside down otherwise...
    // Stucuk: Should be 1- to convert it from DX coords to OpenGL not just -
    Vertices[i].TextureCoord.Y := 1-Vertices[i].TextureCoord.Y;
  end;

  // Seek and read in all the face information
  Seek(F, lumps[kFaces].offset);
  BlockRead(F, Faces[0], numOfFaces*sizeOf(TBSPFace));

  // Seek and read in all the texture information
  Seek(F, lumps[kTextures].offset);
  BlockRead(F, BSPTextures[0], numOfTextures*sizeOf(TBSPTexture));

  // Go through all of the textures
  for I :=0 to numOfTextures-1 do
  begin
    // for some reason known only to ID they dont store file extentions
    if FileExists(BSPTextures[i].TextureName + '.jpg') then
      LoadTexture(BSPTextures[i].TextureName + '.jpg', Textures[i])
    else if FileExists(BSPTextures[i].TextureName + '.tga') then
      LoadTexture(BSPTextures[i].TextureName + '.tga', Textures[i])
  end;
  CloseFile(F);

  result :=TRUE;
end;


{-----------------------------------------------------------}
{--- Renders a face, determined by the passed in index   ---}
{-----------------------------------------------------------}
procedure TQuake3BSP.RenderFace(faceIndex: Integer);
var Face : TBSPFace;
begin
  // Here we grab the face from the index passed in
  Face := Faces[faceIndex];

  glEnable(GL_TEXTURE_2D);                // Turn on texture mapping and bind the face's texture map
  glBindTexture(GL_TEXTURE_2D, textures[Face.textureID]);

  // Draw the face in a triangle face, starting from the starting index
  // to the starting index + the number of vertices.  This is a vertex array function.
  glDrawArrays(GL_TRIANGLE_FAN, Face.startVertIndex, Face.numOfVerts);
end;


{---------------------------------------------------------------------}
{--- Goes through all faces and draws them if type is FACE_POLYGON ---}
{---------------------------------------------------------------------}
procedure TQuake3BSP.RenderLevel(const Pos: TVector3f);
var i : Integer;
begin
  // Give OpenGL our vertices to use for vertex arrays
  glVertexPointer(3, GL_FLOAT, sizeof(TBSPVertex), @Vertices[0].Position);

  // Since we are using vertex arrays, we need to tell OpenGL which texture
  // coordinates to use for each texture pass.  We switch our current texture
  // to the first one, then set our texture coordinates.
  glTexCoordPointer(2, GL_FLOAT, sizeof(TBSPVertex), @Vertices[0].TextureCoord);
  glNormalPointer(GL_FLOAT, sizeof(TBSPVertex), @Vertices[0].Normal);
  // Set our vertex array client states for vertices and texture coordinates
  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_TEXTURE_COORD_ARRAY);

  // Get the number of faces in our level and go through all the faces
  i := numOfFaces;
  while i > 0 do
  begin
    Dec(I);

    // Before drawing this face, make sure it's a normal polygon
    if Faces[i].Facetype = FACE_POLYGON then
      RenderFace(i);
  end;
end;

end.
