// =============================================================================
//   glMatrixHelper
// =============================================================================
// =============================================================================
//   Copyright  2003-2004 by Sascha Willems - http://www.delphigl.de
//   visit the Delphi OpenGL Community - http://www.delphigl.com
// =============================================================================
//   Contents of this file are subject to the GNU Public License (GPL) which can
//   be obtained here : http://opensource.org/licenses/gpl-license.php
//   So only use this file if you fully unterstand that license!!!
// =============================================================================
//   Small unit with some matrix functions usefull for Newton and the GL.
// =============================================================================

unit glMatrixHelper;

interface

uses
 dglOpenGL,glmath;

var
 NullMatrix4f : TMatrix4f = ((0, 0, 0, 0), (0, 0, 0, 0), (0, 0, 0, 0), (0, 0, 0, 0));


procedure Matrix_SetIdentity(var M : TMatrix4f);
procedure Matrix_SetTransform(var M : TMatrix4f; V : TVector4f);
procedure Matrix_SetRotation(var M : TMatrix4f; V : TVector3f);
procedure Matrix_RotateVect(const M : TMatrix4f; var pVect : TVector3f);
function Matrix_MakeYawMatrix(Angle : Single) : TMatrix4f;
function Matrix_MakeRollMatrix(Angle : Single) : TMatrix4f;
function Matrix_Multiply(m1 : TMatrix4f; m2 : TMatrix4f) : TMatrix4f;

function V4(x,y,z : Single) : TVector4f;
function V3(x,y,z : Single) : TVector3f;
function VCross(pV1, pV2 : TVector3f) : TVector3f;

implementation

// =============================================================================
//  V4
// =============================================================================
function V4(x,y,z : Single) : TVector4f;
begin
Result[0] := x;
Result[1] := y;
Result[2] := z;
Result[3] := 0;
end;

// =============================================================================
//  V3
// =============================================================================
function V3(x,y,z : Single) : TVector3f;
begin
Result.X := x;
Result.Y := y;
Result.Z := z;
end;

// =============================================================================
//  VCross
// =============================================================================
function VCross(pV1, pV2 : TVector3f) : TVector3f;
begin
Result.X := (pV1.Y*pV2.z) - (pV1.z*pV2.Y);
Result.Y := (pV1.z*pV2.x) - (pV1.X*pV2.z);
Result.Z := (pV1.X*pV2.y) - (pV1.y*pV2.x);
end;

// =============================================================================
//  Matrix_SetIdentity
// =============================================================================
procedure Matrix_SetIdentity(var M : TMatrix4f);
begin
M[0,0] := 1; M[1,0] := 0; M[2,0] := 0; M[3,0] := 0;
M[0,1] := 0; M[1,1] := 1; M[2,1] := 0; M[3,1] := 0;
M[0,2] := 0; M[1,2] := 0; M[2,2] := 1; M[3,2] := 0;
M[0,3] := 0; M[1,3] := 0; M[2,3] := 0; M[3,3] := 1;
end;

// =============================================================================
//  Matrix_SetTransform
// =============================================================================
procedure Matrix_SetTransform(var M : TMatrix4f; V : TVector4f);
begin
M[3][0] := V[0];
M[3][1] := V[1];
M[3][2] := V[2];
end;

// =============================================================================
//  Matrix_SetRotation
// =============================================================================
procedure Matrix_SetRotation(var M : TMatrix4f; V : TVector3f);
var
 cr , sr , cp , sp , cy , sy , srsp , crsp : single;
begin
V.x := V.x * 180/PI;
V.y := V.y * 180/PI;
V.z := V.z * 180/PI;

cr := cos(V.x);
sr := sin(V.x);
cp := cos(V.y);
sp := sin(V.y);
cy := cos(V.z);
sy := sin(V.z);

M[0,0] := cp*cy;
M[1,0] := cp*sy;
M[2,0] := -sp;
if M[2,0] = -0 then
 M[2,0] := 0;

srsp := sr*sp;
crsp := cr*sp;

M[0,1] := srsp*cy-cr*sy ;
M[1,1] := srsp*sy+cr*cy ;
M[2,1] := sr*cp ;
M[0,2] := crsp*cy+sr*sy ;
M[1,2] := crsp*sy-sr*cy ;
M[2,2] := cr*cp ;
end;

// =============================================================================
//  Matrix_RotateVect
// =============================================================================
procedure Matrix_RotateVect(const M : TMatrix4f; var pVect : TVector3f);
var
 vec : array [0..2] of single;
begin
vec[0] := pVect.x*M[0,0] + pVect.y*M[1,0] + pVect.z*M[2,0];
vec[1] := pVect.x*M[0,1] + pVect.y*M[1,1] + pVect.z*M[2,1];
vec[2] := pVect.x*M[0,2] + pVect.y*M[1,2] + pVect.z*M[2,2];
//vec[0] := pVect[0]*M[0,0] + pVect[1]*M[0,1] + pVect[2]*M[0,2];
//vec[1] := pVect[0]*M[1,0] + pVect[1]*M[1,1] + pVect[2]*M[1,2];
//vec[2] := pVect[0]*M[2,0] + pVect[1]*M[2,1] + pVect[2]*M[2,2];
pVect.x := vec[0];
pVect.y := vec[1];
pVect.z := vec[2];
end;

// =============================================================================
//  Matrix_MakeYawMatrix
// =============================================================================
function Matrix_MakeYawMatrix(Angle : Single) : TMatrix4f;
var
 CA : Single;
 SA : Single;
 M  : TMatrix4f;
begin
SA := Sin(Angle);
CA := Cos(Angle);
M[0,0] := CA; M[1,0] := 0; M[2,0] := -SA; M[3,0] := 0;
M[0,1] := 0;  M[1,1] := 1; M[2,1] := 0;   M[3,1] := 0;
M[0,2] := SA; M[1,2] := 0; M[2,2] := CA;  M[3,2] := 0;
M[0,3] := 0;  M[1,3] := 0; M[2,3] := 0;   M[3,3] := 1;
Result := M;
end;

// =============================================================================
//  Matrix_MakeRollMatrix
// =============================================================================
function Matrix_MakeRollMatrix(Angle : Single) : TMatrix4f;
var
 CA : Single;
 SA : Single;
 M  : TMatrix4f;
begin
SA := Sin(Angle);
CA := Cos(Angle);
M[0,0] := CA;  M[1,0] := SA; M[2,0] := 0; M[3,0] := 0;
M[0,1] := -SA; M[1,1] := CA; M[2,1] := 0; M[3,1] := 0;
M[0,2] := 0;   M[1,2] := 0;  M[2,2] := 1; M[3,2] := 0;
M[0,3] := 0;   M[1,3] := 0;  M[2,3] := 0; M[3,3] := 1;
Result := M;
end;

// =============================================================================
//  Matrix_Multiply
// =============================================================================
function Matrix_Multiply(m1 : TMatrix4f; m2 : TMatrix4f) : TMatrix4f;
var
  r, c, i: Byte;
  t: TMatrix4f;
begin
// Multiply two matrices.
t := NullMatrix4f;
for r := 0 to 3 do
 for c := 0 to 3 do
  for i := 0 to 3 do
   t[r,c] := t[r,c] + (m1[r,i]*m2[i,c]);
Result := t;
end;

end.
